#include "ghead12Periodic.h"

/*****************************************************************************/
/*****************************************************************************/
/******************             sim.c                  ***********************/
/*****************************************************************************/
/*****************************************************************************/


/* The following set of subroutines correspond to the simulation itself. All
   this subroutines share a number of variables which are explained below.
   This variables are capitalized to distinguish them from local variables */

/*  Cond_Pts            Structure that stores the information corresponding
                            to the contioning points retained to krig a
                            particular node
    Kriged_Mean         Result of OK at each simulated location
    Kriged_Variance     Kriging variance at each simulated location
    Kriged_Std          Square root of Kriged_Variance
    Index_Number        Index number that varies randomly between 1 and
                           the total number of nodes
    Krig_Indx[]         Intermediate vector used in the LU decomposition of
                            the kriging matrix
    Krig_Lhs[,]         Matrix that stores the left hand side of the kriging
                        system
    Krig_Nper           Intermediate value used in the LU decomposition of
                            the kriging matrix
    Krig_Rhs[]          Vector that stores the right hand side of the kriging
                        system
    Krig_Vect[]         Intermediate vector used in the LU decomposition of
                            the kriging matrix
    Mod                 Smallest power of 2 that is larger than Total_Nodes.
                            (used to generate the random path)
    Path                Vector of integers running from 1 to Total_Nodes
    Sim_Nodes           A 'ipt' that stores the size of the subarea being
                            simulated
    Total_Nodes         Total number of nodes within simulation sub-area
*/



static int          *Krig_Indx;
static long int     Total_Nodes, Index_Number, Mod, *Path;
static cond_point   Cond_Pts;
static float        *Krig_Rhs, **Krig_Lhs,*Krig_Vect,Krig_Nper;
static ipt          Sim_Nodes;
static float        Kriged_Mean, Kriged_Variance, Kriged_Std;


/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - do_simulation() - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

void do_simulation(fperr,n_nodes,from,to,max_per_octant,nvar,nvar_sim,
                   simulation_var,expected_value,node,node_flag,ntemp,
                   tmp,seed,cova_table,lookup_table,dbg,max_data,isim)
FILE        *fperr;
ipt         from,to,n_nodes;
int         max_per_octant,ntemp,dbg,nvar,nvar_sim,*simulation_var,
            **lookup_table,max_data,isim;
float       expected_value[],***cova_table[];
float       ****node;
short       ***node_flag;
template    *tmp;
long int    *seed;

{
/* This function performs the actual simulation */
/* This version of this subroutine only considers simple co-kriging
   to infer the parameters of the conditional probability distribution
   In the future, ordinary co-kriging will be implemented */

    ipt         point;
    int         j,k,npoints,flag,report,iv,ivar;
    long int    i;


    /* point      stores the coordinates of the point being simulated
       npoints    is the number of conditioning points used by kriging
       report    is an integer used to feed back the user with periodic
                        information about the point that is being simulated
    */
 


    Sim_Nodes.x=to.x-from.x+1;
    Sim_Nodes.y=to.y-from.y+1;
    Sim_Nodes.z=to.z-from.z+1;
    Total_Nodes=(long int)Sim_Nodes.x*(long int)Sim_Nodes.y*(long int)Sim_Nodes.z;
    fprintf(fperr,
         "\nThe total area to be simulated has dimensions %d x %d x %d\n",
                    Sim_Nodes.x, Sim_Nodes.y, Sim_Nodes.z);
/* PERIODIC Begin we must enlarge the x dimension of the simulation area */
    Sim_Nodes.x *= 3;
    Total_Nodes *= 3;
/* PERIODIC End */
    
    allocate_conditioning_data_matrices(nvar,max_data);

    initialize_random_path(fperr,seed);

    report=(log((double)Total_Nodes)-0.5);
    if(report<0) report=0;
    report=pow(E,(double)report);

    printf("Simulation number %d\n",isim);
    for(i=1;i<=Total_Nodes;i++) {
        /* No need to report progress anymore ... year 2013
           if( (i%report) ==1) printf("Now simulating point number %ld\r",i);
         */
                            /* find the coordinates of the point to simulate */

        point=get_next_point(from,i);
/* PERIODIC Begin skip node if not in center cube */
        if(point.x <= n_nodes.x/3 || point.x > 2*n_nodes.x/3 ) continue;
/* PERIODIC End */

                                  /* do nothing if it is a conditioning data with
                                     values for all the variables */
        if(node_flag[point.x][point.y][point.z] == 1) continue;

                      /* find conditioning points within search neighborhood */

        npoints=find_conditioning_points(fperr,n_nodes,point,node,node_flag,
                              ntemp,max_per_octant,tmp,nvar,simulation_var,
                              max_data);
        for(iv=1;iv<=nvar_sim;iv++) {
            ivar=simulation_var[iv];
            if(node[ivar][point.x][point.y][point.z] != UNKNOWN) continue;
            if(dbg>=3) {
              fprintf(fperr, "Point being simulated ---->>> %3d %3d %3d var->%3d\n",
                          point.x,point.y,point.z,ivar);
              fprintf(fperr,"***************************\n\n");
            }

            if(npoints > 0 ) {
	            build_kriging_system(fperr,ivar,point,npoints,cova_table,
	                           lookup_table,dbg);
	            solve_kriging_system(npoints);
            }
	        Kriged_Mean=obtain_SK_estimate(npoints,ivar,expected_value);
	        Kriged_Variance=obtain_SK_kriging_variance
                      (fperr,point,npoints,ivar,cova_table,lookup_table,dbg);
            Kriged_Std=sqrt((double)Kriged_Variance);
            node[ivar][point.x][point.y][point.z]=
                gasdev(seed)*Kriged_Std+Kriged_Mean;
/* PERIODIC Begin */
/* Replicate the point in the other two cubes */
              node[ivar][point.x+n_nodes.x/3][point.y][point.z]=
                 node[ivar][point.x][point.y][point.z];
              node[ivar][point.x-n_nodes.x/3][point.y][point.z]=
                 node[ivar][point.x][point.y][point.z];
/* PERIODIC End   */
            if(dbg>=3) {
                write_out_kriging_weights(fperr,npoints);
                write_out_cpdf(fperr,node[ivar][point.x][point.y][point.z]);
            }
                    /* include just kriged variable in the conditioning points */
            npoints++;
            Cond_Pts.ix[npoints]=point.x;
            Cond_Pts.iy[npoints]=point.y;
            Cond_Pts.iz[npoints]=point.z;
            Cond_Pts.index[npoints]=ivar;
            Cond_Pts.value[npoints]=node[ivar][point.x][point.y][point.z];
        }
        node_flag[point.x][point.y][point.z]=1;

/* PERIODIC Begin  flag the replicates, too */
              node_flag[point.x+n_nodes.x/3][point.y][point.z]=1;
              node_flag[point.x-n_nodes.x/3][point.y][point.z]=1;
/* PERIODIC End   */

    }
     free_memory(nvar,max_data); 
}

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - -  - - - - - - - - - obtain_SK()  - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

float obtain_SK_estimate(npoints,ivar,expected_value)
int npoints,ivar;
float *expected_value;
{
/* the linear simple kriging estimate of the kriging estimate of the
   conditional probability is computed here */

   int i;
   float estimate;

   estimate=expected_value[ivar];
   for(i=1;i<=npoints;i++)
       estimate += Krig_Rhs[i]*(Cond_Pts.value[i]-
                                expected_value[Cond_Pts.index[i]]);

   return estimate;
}

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - free_memory() - - - - - -- - - - - -- - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

void free_memory(nvar,max_data)
int nvar,max_data;
{
     int maxpt;
     maxpt=nvar*max_data;
     free_ivector(Cond_Pts.ix,1,maxpt);
     free_ivector(Cond_Pts.iy,1,maxpt);
     free_ivector(Cond_Pts.iz,1,maxpt);
     free_vector(Cond_Pts.value,1,maxpt);
     free_ivector(Cond_Pts.index,1,maxpt);
     free_matrix(Krig_Lhs,1,maxpt+1,1,maxpt+1);
     free_vector(Krig_Rhs,1,maxpt+1);
     free_ivector(Krig_Indx,1,maxpt+1);
     free_vector(Krig_Vect,1,maxpt+1);
     free_lvector(Path,1L,Total_Nodes);
}

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - allocate_conditioning_data_matrices() - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

void allocate_conditioning_data_matrices(nvar,max_data)
int nvar,max_data;
{
    int maxpt;
    maxpt=nvar*max_data+1;
    Cond_Pts.ix=ivector(1,maxpt);
    Cond_Pts.iy=ivector(1,maxpt);
    Cond_Pts.iz=ivector(1,maxpt);
    Cond_Pts.value=vector(1,maxpt);
    Cond_Pts.index=ivector(1,maxpt);

    Krig_Lhs=matrix(1,maxpt+1,1,maxpt+1);
    Krig_Rhs=vector(1,maxpt+1);
    Krig_Indx=ivector(1,maxpt+1);
    Krig_Vect=vector(1,maxpt+1);

}

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - initialize_random_path()  - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

void initialize_random_path(fperr, seed)
FILE        *fperr;
long         *seed;
{
/* This function allocates Path and creates a random permutation of integers 
   between 1 and Total_Nodes */

	long dummy, i, j;

	Path = lvector(1L, Total_Nodes);
	for (i=1;i<=Total_Nodes;i++) 
		Path[i]=i;

    for (i=1;i<=Total_Nodes;i++) {
		j = (long) (rand0(seed)*Total_Nodes+1);
		if (j < 1 || j > Total_Nodes) 
			error ("Error in initialize_random_path");
		dummy=Path[i];
		Path[i]=Path[j];
		Path[j]=dummy;
	}
	
	Index_Number=1L; /* legcay code to keep using Index_Number, now Index_Number
	                   runs from 1 to Total_Nodes sequentially */

    fprintf(fperr,
       "\nA random permutation of %ld integers has been created:\n",Total_Nodes);
    fprintf(fperr,"First few values are %ld %ld %ld %ld %ld\n\n",Path[1],Path[2],Path[3],Path[4],Path[5]);

}

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - -  get_next_point() - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

ipt get_next_point(from,i)
ipt         from;
long        i;
{
/* This function returns the coordinates of the next point to be simulated.
   Each node in the simulation sub-area can be uniquely determined by an index
   number between 1 and Total_Nodes according to the following system:

        index           coordinates realtive to (from.x, from.y, from.z)
        -----           ------------------------------------------------
          0                 (0,0,0)
          1                 (0,0,1)
          .                    .                    z increments fastest
          .                    .
        Sim_Nodes.z-1       (0,0,Sim_Nodes.z)
        Sim_Nodes.z         (0,1,0)
        Sim_Nodes.z+1       (0,1,1)
          .                    .                    y increments next
          .                    .
Sim_Nodes.y*Sim_Nodes.z-1   (0,Sim_Nodes.y,Sim_Nodes.z)
Sim_Nodes.y*Sim_Nodes.z     (1,0,0)
          .                    .                    x increments slowest
          .                    .
       Total_Nodes-1          (Sim_Nodes.x,Sim_Nodes.y,Sim_Nodes.z)
*/


    ipt             point;
    long int        dummy,random_index,nyz;


	random_index = Path[i];
    random_index--; /* index goes from 0 to Total_Nodes-1 */
  
             /* calculate coordinates relative to sub-area being simulated */

    nyz     = (long int)Sim_Nodes.y*(long int)Sim_Nodes.z;
    point.x = random_index/nyz;
    dummy   = random_index-point.x*nyz;
    point.y = dummy/Sim_Nodes.z;
    point.z = dummy-point.y*Sim_Nodes.z;

                                          /* calculate true grid coordinates */

    point.x += from.x;
    point.y += from.y;
    point.z += from.z;

    return point;
}

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - find_conditioning_points()  - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

int find_conditioning_points(fperr,n_nodes,point,node,node_flag,ntemp,
                               max_per_octant,tmp,nvar,simulation_var,max_data)
FILE            *fperr;
ipt             point,n_nodes;
float           ****node;
short           ***node_flag;
int             *simulation_var;
int             ntemp, max_per_octant,nvar,max_data;
template        *tmp;
{
/* For the node with coordinates stored in 'point', this function will find
   the nodes within the search neighborhood that have already being simulated
   or that are initial data. To do this the function makes use of the template
   previously calculated in which the relative coordinates of all possible
   points have being stored in order of distance starting with the closest
   point. There is a maximum number of points per octant that can be used,
   so the maximum number points to be used is 8*max_per_octant. The function
   returns the number of conditioning points found. */

    int     i,j,jx,jy,jz,ioct,points_per_octant[8],point_counter,iv;

                             /* initialize octant_counters and point counter */

    if(max_per_octant>0)
      for(i=0;i<8;i++)
        points_per_octant[i]=0;

    point_counter=0;


    /* retrieve the points from the template in order of decreasing distance */

    for(i=1;i<=ntemp && point_counter<max_data;i++) {
        j=tmp->index[i];
                                              /* obtain absolute coordinates */
        jx=tmp->irx[j]+point.x;
        jy=tmp->iry[j]+point.y;
        jz=tmp->irz[j]+point.z;

   /* if the point is outside the grid or has not been simulated yet discard */

        if (jx<1 || jy<1 || jz<1 ||
            jx>n_nodes.x || jy>n_nodes.y || jz>n_nodes.z ||
            node_flag[jx][jy][jz] == -1 ) continue;

                        /* if there are already more than max_per_octant points
                           in the octant to which the point belongs, discard
                           otherwise move the point to the Cond_Pts structure
                           and increase octant counter and point counter */

        if(max_per_octant>0) {
           ioct=tmp->octant[j];
           if(points_per_octant[ioct] >= max_per_octant ) {
              continue;
           }
           else points_per_octant[ioct]++;
        }

        for (iv=1;iv<=nvar;iv++)  {

            if(node[iv][jx][jy][jz]==UNKNOWN) continue;
            point_counter++;

            Cond_Pts.ix[point_counter]=jx;
            Cond_Pts.iy[point_counter]=jy;
            Cond_Pts.iz[point_counter]=jz;
            Cond_Pts.value[point_counter]=node[iv][jx][jy][jz];
            Cond_Pts.index[point_counter]=iv;
        }


        /* check if there is any octant that is not complete */

        if(max_per_octant > 0) {
           for(ioct=0;ioct<8;ioct++) {
              if (points_per_octant[ioct] <= max_per_octant) break;
           }

           if(ioct<8) continue;
           else break;
        }
    }

    return point_counter;
}

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - build_kriging_system()  - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

void    build_kriging_system(fperr,ivar,point,npoints,cova_table,lookup_table,
        dbg)
int     npoints,dbg,**lookup_table,ivar;
float   ****cova_table;
FILE    *fperr;
ipt     point;
{
/* the left hand side matrix and the right hand side vector of the kriging
   system are built here using the covariance look up tables previously
   intialized */

    int i,j,dx,dy,dz,acov;

    for(i=1;i<=npoints;i++) {
        for(j=1;j<=i;j++) {
            dx = Cond_Pts.ix[i]-Cond_Pts.ix[j];
            dy = Cond_Pts.iy[i]-Cond_Pts.iy[j];
            dz = Cond_Pts.iz[i]-Cond_Pts.iz[j];
            acov=lookup_table[Cond_Pts.index[i]][Cond_Pts.index[j]];
            Krig_Lhs[i][j]= dx < 0 ? cova_table[acov][-dx][-dy][-dz] :
                                      cova_table[acov][dx][dy][dz];
            Krig_Lhs[j][i]=Krig_Lhs[i][j];
        }
        dx=Cond_Pts.ix[i]-point.x;
        dy=Cond_Pts.iy[i]-point.y;
        dz=Cond_Pts.iz[i]-point.z;
        acov=lookup_table[ivar][Cond_Pts.index[i]];
        Krig_Rhs[i]= dx < 0 ? cova_table[acov][-dx][-dy][-dz] :
                                      cova_table[acov][dx][dy][dz];
    }

    if(dbg >= 3) {
        fprintf(fperr,"Left Hand Side Kriging Matrix\n");
        for (i=1;i<=npoints; i++) {
            for (j=1;j<=i;j++) {
                fprintf(fperr,"%8.3f",Krig_Lhs[i][j]);
            }
            fprintf(fperr,"\n");
        }
        fprintf(fperr,"\n Right Hand Side Kriging Column Vector\n");
        for(i=1;i<=npoints;i++) {
            fprintf(fperr,"%8.3f",Krig_Rhs[i]);
        }
        fprintf(fperr,"\n\n");
    }
}

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - solve_kriging_system()  - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

void solve_kriging_system(nequations)
int     nequations;
{
/* solution of the kriging system by means of the LU decomposition */

    ludcmp(Krig_Lhs,nequations,Krig_Indx,&Krig_Nper);
    lubksb(Krig_Lhs,nequations,Krig_Indx,Krig_Rhs);
}

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - -obtain_SK_kriging_variance() - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

float obtain_SK_kriging_variance(fperr,point,npoints,ivar,cova_table,
                     lookup_table,dbg)
FILE   *fperr;
int    npoints,ivar,dbg;
ipt    point;
float  ****cova_table;
int    **lookup_table;
{
     float variance;
     int dx,dy,dz,i,acov;

     acov=lookup_table[ivar][ivar];
     variance=cova_table[acov][0][0][0];
     for(i=1;i<=npoints;i++){
        dx=Cond_Pts.ix[i]-point.x;
        dy=Cond_Pts.iy[i]-point.y;
        dz=Cond_Pts.iz[i]-point.z;
        acov=lookup_table[ivar][Cond_Pts.index[i]];

        variance -=  (dx < 0 ? cova_table[acov][-dx][-dy][-dz] :
                               cova_table[acov][dx][dy][dz]) * Krig_Rhs[i];
     }

     return variance;
}


/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - write_out_kriging_weights() - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

void write_out_kriging_weights(fperr,npoints)
FILE *fperr;
int  npoints;
{
/* this function is used only for debugging purposes. It prints relevant
   information to the solution of the kriging systems */

    int i;

    fprintf(fperr,"A total of ___ %d ___ points were found\n\n",npoints);
    fprintf(fperr,"       ix        iy        iz       var     value    weight\n");
    for(i=1;i<=npoints;i++) {
        fprintf(fperr,
                    "%9d %9d %9d %9d %9.2f %10.3f\n",Cond_Pts.ix[i],Cond_Pts.iy[i],
                    Cond_Pts.iz[i],Cond_Pts.index[i],Cond_Pts.value[i], Krig_Rhs[i] );
    }
    fprintf(fperr,"\n");
}

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - -  write_out_cpdf() - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

void write_out_cpdf(FILE *fperr,float value)
{
/* this function is only used for debugging purposes, it prints the estimate
   of the conditional probability distribution function stored in Cpdf */

    fprintf(fperr,"SK estimates           \n");
    fprintf(fperr,"------------           \n");
    fprintf(fperr,"Mean         Variance  \n");
    fprintf(fperr,"%10.3f %10.3f\n",Kriged_Mean,Kriged_Variance);
    fprintf(fperr,"Random drawing --> %10.3f\n",value);
}


/* the following subroutines are based in those from the book Numerical
   Recipes in C, by Press, Flannery, Teukolsky and Vetterling, Cambridge
   University Press. They implement the solution of a linear system by the LU
   decomposition method */

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - ludcmp()  - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/


void ludcmp(a,n,indx,d)
int n,*indx;
float **a,*d;
{
/* this is the LU decomposition itself */

    int i,imax,j,k;
    float big,dum,sum,temp;

/* since this subroutine is called many times, the allocation of the vector
   'vv' is done outside. This vector was called Krig_Vect and is #define'd here
   as 'vv'. */

#define vv Krig_Vect

    *d=1.0;
    for (i=1;i<=n;i++) {
        big=0.0;
        for (j=1;j<=n;j++)
            if ((temp=fabs(a[i][j])) > big) big=temp;
        if (big == 0.0) error("Singular matrix in routine LUDCMP");
        vv[i]=1.0/big;
    }
    for (j=1;j<=n;j++) {
        for (i=1;i<j;i++) {
            sum=a[i][j];
            for (k=1;k<i;k++) sum -= a[i][k]*a[k][j];
            a[i][j]=sum;
        }
        big=0.0;
        for (i=j;i<=n;i++) {
            sum=a[i][j];
            for (k=1;k<j;k++)
                sum -= a[i][k]*a[k][j];
            a[i][j]=sum;
            if ( (dum=vv[i]*fabs(sum)) >= big) {
                big=dum;
                imax=i;
            }
        }
        if (j != imax) {
            for (k=1;k<=n;k++) {
                dum=a[imax][k];
                a[imax][k]=a[j][k];
                a[j][k]=dum;
            }
            *d = -(*d);
            vv[imax]=vv[j];
        }
        indx[j]=imax;
        if (a[j][j] == 0.0) a[j][j]=TINY;
        if (j != n) {
            dum=1.0/(a[j][j]);
            for (i=j+1;i<=n;i++) a[i][j] *= dum;
        }
    }
#undef vv
}

#undef TINY

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - lubksb()  - - - - - - - - - - - - - - - -*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

void lubksb(a,n,indx,b)
float **a,b[];
int n,*indx;
{
/* this is the solution of the system LUx=b, in which LU is the LU
   decomposition of matrix 'a' */

    int i,ii=0,ip,j;
    float sum;

    for (i=1;i<=n;i++) {
        ip=indx[i];
        sum=b[ip];
        b[ip]=b[i];
        if (ii)
            for (j=ii;j<=i-1;j++) sum -= a[i][j]*b[j];
        else if (sum) ii=i;
        b[i]=sum;
    }
    for (i=n;i>=1;i--) {
        sum=b[i];
        for (j=i+1;j<=n;j++) sum -= a[i][j]*b[j];
        b[i]=sum/a[i][i];
    }
}


float gasdev(idum)
long *idum;

{
      static int iset=0;
      static float gset;
      float fac,r,v1,v2;

      if(iset == 0) {
          do {
             v1=2.0*rand0(idum)-1.0;
             v2=2.0*rand0(idum)-1.0;
             r=v1*v1+v2*v2;
          } while (r >= 1.0);
          fac=sqrt(-2.0*log(r)/r);
          gset=v1*fac;
          iset=1;
          return v2*fac;
      } else {
          iset=0;
          return gset;
      }
}
